/*    Filename: tc-1.c
 *      Author: Sripathi Kodi <sripathik@in.ibm.com>
 * Description: Check if CLOCK_THREAD_CPUTIME_ID is working correctly.
 * - Creates a worker thread that tightly binds to the CPU by running a loop
 * - Creates a sleeper thread that just sleeps :)
 * The thread time returned by the CLOCK_THREAD_CPUTIME_ID must be non-zero for
 * the worker thread but almost negligible for the sleeper thread (as it only
 * sleeps and does not hog the CPU).
 *
 * Compilation: gcc tc-1.c -lpthread -lrt -o tc-1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (C) IBM Corporation, 2006
 *
 * 2006-Nov-14:  Initial version by Sripathi Kodi <sripathik@in.ibm.com>
 */

#include <stdio.h>
#include <pthread.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

/* Just spend some time on the CPU */
void work(void)
{
	unsigned int i=0;
	for (i=0; i<2147483600; i++) {
		if ((i == i+1) || (i == i-1))
			printf("Hey!\n");
	}
}

#define GET_MY_THREAD_TIME(ts)						\
	if ((clock_gettime (CLOCK_THREAD_CPUTIME_ID, ts)) < 0) {	\
		perror("clock_gettime: CLOCK_THREAD_CPUTIME_ID: ");	\
		exit(1);						\
	}

void *workerthread(void *arg)
{
	struct timespec *ts = (struct timespec *)arg;

#ifdef DEBUG
	printf("Worker thread working\n");
#endif
	work();

	GET_MY_THREAD_TIME(ts);

#ifdef DEBUG
	printf("workerthread: AFTER WORK: tv_sec = %ld, tv_nsec = %ld\n", ts->tv_sec, ts->tv_nsec);
#endif
	return 0;
}


void *sleeperthread(void *arg)
{
	struct timespec *ts = (struct timespec *)arg;

#ifdef DEBUG
	printf("Sleeper thread sleeping\n");
#endif

	sleep(5);

	GET_MY_THREAD_TIME(ts);

#ifdef DEBUG
	printf("sleeperthread: AFTER SLEEP: tv_sec = %ld, tv_nsec = %ld\n", ts->tv_sec, ts->tv_nsec);
#endif
	return 0;
}

int main()
{
	pthread_t thread1, thread2;
	struct timespec ts1, ts2;
	int retval = 0;

	if ((pthread_create (&thread1, NULL, workerthread, (void *)&ts1)) < 0 ) {
		perror("pthread_create: ");
		exit(1);
	}
	printf("\nworker thread created\n");
	if ((pthread_create (&thread2, NULL, sleeperthread, (void *)&ts2)) < 0 ) {
		perror("pthread_create: ");
		exit(1);
	}
	printf("sleeper thread created\n");
	printf("Please wait...\n");

	pthread_join(thread1, NULL);
	pthread_join(thread2, NULL);
	printf("workerthread spent %ld seconds, %ld nanoseconds\n", ts1.tv_sec, ts1.tv_nsec);
	printf("sleeperthread spent %ld seconds, %ld nanoseconds\n", ts2.tv_sec, ts2.tv_nsec);

	if (ts1.tv_sec == 0) {
		printf("\n!!! FAIL !!! worker thread clock values seem to be wrong\n");
		retval = 1;
	}
	if (ts2.tv_sec != 0) {
		printf("\n!!! FAIL !!! sleeper thread clock values seem to be wrong\n");
		retval = 1;
	}
	printf("\n Test %s\n\n", retval==0 ? "PASSED" : "FAILED");
	return retval;
}
